@@ -660,13 +660,13 @@ static void _c2s_component_presence(c2s_t c2s, nad_t nad) {
         if(xhash_iter_first(c2s->sessions))
             do {
                 xhv.sess_val = &sess;
-                xhash_iter_get(c2s->sessions, NULL, xhv.val);
+                xhash_iter_get(c2s->sessions, NULL, NULL, xhv.val);
 
                 if(sess->resources != NULL && strcmp(sess->resources->jid->domain, from) == 0) {
                     log_debug(ZONE, "killing session %s", jid_user(sess->resources->jid));
 
                     sess->active = 0;
-                    sx_close(sess->s);
+                    if(sess->s) sx_close(sess->s);
                 }
             } while(xhash_iter_next(c2s->sessions));
 
@@ -680,7 +680,7 @@ int c2s_router_sx_callback(sx_t s, sx_event_t e, void *data, void *arg) {
     sx_error_t *sxe;
     nad_t nad;
     int len, elem, from, c2sid, smid, action, id, ns, attr, scan, replaced;
-    char skey[10];
+    char skey[44];
     sess_t sess;
     bres_t bres, ires;
 
@@ -760,8 +760,8 @@ int c2s_router_sx_callback(sx_t s, sx_event_t e, void *data, void *arg) {
         case event_OPEN:
             log_write(c2s->log, LOG_NOTICE, "connection to router established");
 
-            /* reset connection attempts counter */
-            c2s->retry_left = c2s->retry_init;
+            /* set connection attempts counter */
+            c2s->retry_left = c2s->retry_lost;
 
             nad = nad_new();
             ns = nad_add_namespace(nad, uri_COMPONENT, NULL);
@@ -863,15 +863,19 @@ int c2s_router_sx_callback(sx_t s, sx_event_t e, void *data, void *arg) {
                 }
 
 #ifdef HAVE_SSL
-                if(c2s->server_fd == NULL && c2s->server_ssl_fd == NULL) {
+                if(c2s->server_fd == NULL && c2s->server_ssl_fd == NULL && c2s->pbx_pipe == NULL) {
                     log_write(c2s->log, LOG_ERR, "both normal and SSL ports are disabled, nothing to do!");
 #else
-                if(c2s->server_fd == NULL) {
+                if(c2s->server_fd == NULL && c2s->pbx_pipe == NULL) {
                     log_write(c2s->log, LOG_ERR, "server port is disabled, nothing to do!");
 #endif
                     exit(1);
                 }
 
+                /* open PBX integration FIFO */
+                if(c2s->pbx_pipe != NULL)
+                    c2s_pbx_init(c2s);
+
                 /* we're online */
                 c2s->online = c2s->started = 1;
                 log_write(c2s->log, LOG_NOTICE, "ready for connections", c2s->id);
@@ -928,20 +932,68 @@ int c2s_router_sx_callback(sx_t s, sx_event_t e, void *data, void *arg) {
                 nad_free(nad);
                 return 0;
             }
-            snprintf(skey, 10, "%.*s", NAD_AVAL_L(nad, c2sid), NAD_AVAL(nad, c2sid));
+            snprintf(skey, 44, "%.*s", NAD_AVAL_L(nad, c2sid), NAD_AVAL(nad, c2sid));
 
             /* find the session, quietly drop if we don't have it */
             sess = xhash_get(c2s->sessions, skey);
             if(sess == NULL) {
-                /* !!! might want to send a stop to the sm; maybe it thinks we're still here? */
+                /* if we get this, the SM probably thinks the session is still active
+                 * so we need to tell SM to free it up */
                 log_debug(ZONE, "no session for %s", skey);
 
+                /* check if it's a started action; otherwise we could end up in an infinite loop
+                 * trying to tell SM to close in response to errors */
+                action = nad_find_attr(nad, 1, -1, "action", NULL);
+                if(action >= 0 && NAD_AVAL_L(nad, action) == 7 && strncmp("started", NAD_AVAL(nad, action), 7) == 0) {
+					int target;
+					bres_t tres;
+					sess_t tsess;
+
+                    log_write(c2s->log, LOG_NOTICE, "session %s does not exist; telling sm to close", skey);
+
+                    /* we don't have a session and we don't have a resource; we need to forge them both
+                     * to get SM to close stuff */
+                    target = nad_find_attr(nad, 1, -1, "target", NULL);
+                    smid = nad_find_attr(nad, 1, ns, "sm", NULL);
+                    if(target < 0 || smid < 0) {
+                        char *buf;
+                        int len;
+                        nad_print(nad, 0, &buf, &len);
+                        log_write(c2s->log, LOG_NOTICE, "sm sent an invalid start packet: %.*s", len, buf );
+                        nad_free(nad);
+                        return 0;
+                    }
+
+                    /* build temporary resource to close session for */
+                    tres = NULL;
+                    tres = (bres_t) calloc(1, sizeof(struct bres_st));
+                    tres->jid = jid_new(NAD_AVAL(nad, target), NAD_AVAL_L(nad, target));
+
+                    strncpy(tres->c2s_id, skey, sizeof(tres->c2s_id));
+                    snprintf(tres->sm_id, sizeof(tres->sm_id), "%.*s", NAD_AVAL_L(nad, smid), NAD_AVAL(nad, smid));
+
+                    /* make a temporary session */
+                    tsess = (sess_t) calloc(1, sizeof(struct sess_st));
+                    tsess->c2s = c2s;
+                    tsess->result = nad_new();
+                    strncpy(tsess->skey, skey, sizeof(tsess->skey));
+
+                    /* end a session with the sm */
+                    sm_end(tsess, tres);
+
+                    /* free our temporary messes */
+                    nad_free(tsess->result);
+                    jid_free(tres->jid); //TODO will this crash?
+                    free(tsess);
+                    free(tres);
+                }
+
                 nad_free(nad);
                 return 0;
             }
 
             /* if they're pre-stream, then this is leftovers from a previous session */
-            if(sess->s->state < state_STREAM) {
+            if(sess->s && sess->s->state < state_STREAM) {
                 log_debug(ZONE, "session %s is pre-stream", skey);
 
                 nad_free(nad);
@@ -970,8 +1022,10 @@ int c2s_router_sx_callback(sx_t s, sx_event_t e, void *data, void *arg) {
             if(nad_find_attr(nad, 0, -1, "error", NULL) >= 0) {
                 log_debug(ZONE, "routing error");
 
-                sx_error(sess->s, stream_err_INTERNAL_SERVER_ERROR, "internal server error");
-                sx_close(sess->s);
+                if(sess->s) {
+                    sx_error(sess->s, stream_err_INTERNAL_SERVER_ERROR, "internal server error");
+                    sx_close(sess->s);
+                }
 
                 nad_free(nad);
                 return 0;
@@ -1006,7 +1060,7 @@ int c2s_router_sx_callback(sx_t s, sx_event_t e, void *data, void *arg) {
                 tres = (bres_t) calloc(1, sizeof(struct bres_st));
                 tres->jid = jid;
                 sprintf(tres->c2s_id, "%d", sess->s->tag);
-                snprintf(tres->sm_id, 41, "%.*s", NAD_AVAL_L(nad, smid), NAD_AVAL(nad, smid));
+                snprintf(tres->sm_id, sizeof(tres->sm_id), "%.*s", NAD_AVAL_L(nad, smid), NAD_AVAL(nad, smid));
 
                 if(sess->resources) {
                     log_debug(ZONE, "expected packet from sm session %s, but got one from %.*s, ending sm session", sess->resources->sm_id, NAD_AVAL_L(nad, smid), NAD_AVAL(nad, smid));
@@ -1027,14 +1081,6 @@ int c2s_router_sx_callback(sx_t s, sx_event_t e, void *data, void *arg) {
                 return 0;
             }
 
-            /* it has to have come from the session manager */
-            from = nad_find_attr(nad, 0, -1, "from", NULL);
-            if(!sess->s->req_to || strlen(sess->s->req_to) != NAD_AVAL_L(nad, from) || strncmp(sess->s->req_to, NAD_AVAL(nad, from), NAD_AVAL_L(nad, from)) != 0) {
-                log_debug(ZONE, "packet from '%.*s' for %s, but they're not the sm for this sess", NAD_AVAL_L(nad, from), NAD_AVAL(nad, from), skey);
-                nad_free(nad);
-                return 0;
-            }
-
             /* session control packets */
             if(NAD_ENS(nad, 1) == ns && action >= 0) {
                 /* end responses */
@@ -1063,8 +1109,8 @@ int c2s_router_sx_callback(sx_t s, sx_event_t e, void *data, void *arg) {
                         if(replaced)
                             sx_error(sess->s, stream_err_CONFLICT, NULL);
 
-                        /* close them */
-                        sx_close(sess->s);
+                        /* close the stream if there is one */
+                        if(sess->s) sx_close(sess->s);
 
                         nad_free(nad);
                         return 0;
@@ -1161,18 +1207,27 @@ int c2s_router_sx_callback(sx_t s, sx_event_t e, void *data, void *arg) {
 
                     /* copy the sm id */
                     if(smid >= 0)
-                        snprintf(bres->sm_id, 41, "%.*s", NAD_AVAL_L(nad, smid), NAD_AVAL(nad, smid));
+                        snprintf(bres->sm_id, sizeof(bres->sm_id), "%.*s", NAD_AVAL_L(nad, smid), NAD_AVAL(nad, smid));
+
+                    /* and remember the SM that services us */
+                    from = nad_find_attr(nad, 0, -1, "from", NULL);
+                    sess->smcomp = malloc(NAD_AVAL_L(nad, from) + 1);
+                    snprintf(sess->smcomp, NAD_AVAL_L(nad, from) + 1, "%.*s", NAD_AVAL_L(nad, from), NAD_AVAL(nad, from));
 
                     nad_free(nad);
 
                     /* bring them online, old-skool */
-                    if(!sess->sasl_authd) {
+                    if(!sess->sasl_authd && sess->s) {
                         sx_auth(sess->s, "traditional", jid_full(bres->jid));
                         return 0;
                     }
 
-                    /* return the auth result to the client */
-                    if(sess->result) sx_nad_write(sess->s, sess->result);
+                    if(sess->result) {
+                        /* return the auth result to the client */
+                        if(sess->s) sx_nad_write(sess->s, sess->result);
+                        /* or follow-up the session creation with cached presence packet */
+                        else sm_packet(sess, bres, sess->result);
+                    }
                     sess->result = NULL;
 
                     /* we're good to go */
@@ -1193,7 +1248,7 @@ int c2s_router_sx_callback(sx_t s, sx_event_t e, void *data, void *arg) {
 
             /* client packets */
             if(NAD_NURI_L(nad, NAD_ENS(nad, 1)) == strlen(uri_CLIENT) && strncmp(uri_CLIENT, NAD_NURI(nad, NAD_ENS(nad, 1)), strlen(uri_CLIENT)) == 0) {
-                if(!sess->active) {
+                if(!sess->active || !sess->s) {
                     /* its a strange world .. */
                     nad_free(nad);
                     return 0;
@@ -1201,9 +1256,11 @@ int c2s_router_sx_callback(sx_t s, sx_event_t e, void *data, void *arg) {
 
                 /* sm is bouncing something */
                 if(nad_find_attr(nad, 1, ns, "failed", NULL) >= 0) {
-                    /* there's really no graceful way to handle this */
-                    sx_error(s, stream_err_INTERNAL_SERVER_ERROR, "session manager failed control action");
-                    sx_close(s);
+                    if(s) {
+                        /* there's really no graceful way to handle this */
+                        sx_error(s, stream_err_INTERNAL_SERVER_ERROR, "session manager failed control action");
+                        sx_close(s);
+                    }
 
                     nad_free(nad);
                     return 0;
