@@ -123,6 +123,8 @@ static void _c2s_config_expand(c2s_t c2s)
 
     c2s->local_pemfile = config_get_one(c2s->config, "local.pemfile", 0);
 
+    c2s->local_cachain = config_get_one(c2s->config, "local.cachain", 0);
+
     c2s->local_verify_mode = j_atoi(config_get_one(c2s->config, "local.verify-mode", 0), 0);
 
     c2s->local_ssl_port = j_atoi(config_get_one(c2s->config, "local.ssl-port", 0), 0);
@@ -137,6 +139,8 @@ static void _c2s_config_expand(c2s_t c2s)
     c2s->io_check_idle = j_atoi(config_get_one(c2s->config, "io.check.idle", 0), 0);
     c2s->io_check_keepalive = j_atoi(config_get_one(c2s->config, "io.check.keepalive", 0), 0);
 
+    c2s->pbx_pipe = config_get_one(c2s->config, "pbx.pipe", 0);
+
     c2s->ar_module_name = config_get_one(c2s->config, "authreg.module", 0);
 
     if(config_get(c2s->config, "authreg.mechanisms.traditional.plain") != NULL) c2s->ar_mechanisms |= AR_MECH_TRAD_PLAIN;
@@ -241,8 +245,8 @@ static void _c2s_hosts_expand(c2s_t c2s)
 
     elem = config_get(c2s->config, "local.id");
     if(!elem) {
-        log_write(c2s->log, LOG_ERR, "no local.id configured, aborting");
-        exit(1);
+        log_write(c2s->log, LOG_NOTICE, "no local.id configured - skipping local domains configuration");
+        return;
     }
     for(i = 0; i < elem->nvalues; i++) {
         host_t host = (host_t) pmalloco(xhash_pool(c2s->hosts), sizeof(struct host_st));
@@ -325,7 +329,7 @@ static void _c2s_hosts_expand(c2s_t c2s)
 static int _c2s_router_connect(c2s_t c2s) {
     log_write(c2s->log, LOG_NOTICE, "attempting connection to router at %s, port=%d", c2s->router_ip, c2s->router_port);
 
-    c2s->fd = mio_connect(c2s->mio, c2s->router_port, c2s->router_ip, c2s_router_mio_callback, (void *) c2s);
+    c2s->fd = mio_connect(c2s->mio, c2s->router_port, c2s->router_ip, NULL, c2s_router_mio_callback, (void *) c2s);
     if(c2s->fd == NULL) {
         if(errno == ECONNREFUSED)
             c2s_lost_router = 1;
@@ -520,9 +524,9 @@ static void _c2s_time_checks(c2s_t c2s) {
     if(xhash_iter_first(c2s->sessions))
         do {
             xhv.sess_val = &sess;
-            xhash_iter_get(c2s->sessions, NULL, xhv.val);
+            xhash_iter_get(c2s->sessions, NULL, NULL, xhv.val);
 
-            if(c2s->io_check_idle > 0 && now > sess->last_activity + c2s->io_check_idle) {
+            if(c2s->io_check_idle > 0 && sess->s && now > sess->last_activity + c2s->io_check_idle) {
                 log_write(c2s->log, LOG_NOTICE, "[%d] [%s, port=%d] timed out", sess->fd->fd, sess->ip, sess->port);
 
                 sx_error(sess->s, stream_err_HOST_GONE, "connection timed out");
@@ -645,7 +649,9 @@ JABBER_MAIN("jabberd2c2s", "Jabber 2 C2S", "Jabber Open Source Server: Client to
     _c2s_pidfile(c2s);
 
     if(c2s->ar_module_name == NULL) {
-        log_write(c2s->log, LOG_ERR, "no authreg module specified in config file");
+        log_write(c2s->log, LOG_NOTICE, "no authreg module specified in config file");
+    }
+    else if((c2s->ar = authreg_init(c2s, c2s->ar_module_name)) == NULL) {
         access_free(c2s->access);
         config_free(c2s->config);
         log_free(c2s->log);
@@ -653,14 +659,6 @@ JABBER_MAIN("jabberd2c2s", "Jabber 2 C2S", "Jabber Open Source Server: Client to
         exit(1);
     }
 
-    if((c2s->ar = authreg_init(c2s, c2s->ar_module_name)) == NULL) {
-        access_free(c2s->access);
-        config_free(c2s->config);
-        log_free(c2s->log);
-        free(c2s);
-        exit(1);
-        }
-
     c2s->sessions = xhash_new(1023);
 
     c2s->conn_rates = xhash_new(101);
@@ -674,7 +672,7 @@ JABBER_MAIN("jabberd2c2s", "Jabber 2 C2S", "Jabber Open Source Server: Client to
 #ifdef HAVE_SSL
     /* get the ssl context up and running */
     if(c2s->local_pemfile != NULL) {
-        c2s->sx_ssl = sx_env_plugin(c2s->sx_env, sx_ssl_init, NULL, c2s->local_pemfile, NULL, c2s->local_verify_mode);
+        c2s->sx_ssl = sx_env_plugin(c2s->sx_env, sx_ssl_init, NULL, c2s->local_pemfile, c2s->local_cachain, c2s->local_verify_mode);
         if(c2s->sx_ssl == NULL) {
             log_write(c2s->log, LOG_ERR, "failed to load local SSL pemfile, SSL will not be available to clients");
             c2s->local_pemfile = NULL;
@@ -716,6 +714,10 @@ JABBER_MAIN("jabberd2c2s", "Jabber 2 C2S", "Jabber Open Source Server: Client to
     sx_env_plugin(c2s->sx_env, bind_init, c2s);
 
     c2s->mio = mio_new(c2s->io_max_fds);
+    if(c2s->mio == NULL) {
+        log_write(c2s->log, LOG_ERR, "failed to create MIO, aborting");
+        exit(1);
+    }
 
     /* hosts mapping */
     c2s->hosts = xhash_new(1021);
@@ -745,6 +747,7 @@ JABBER_MAIN("jabberd2c2s", "Jabber 2 C2S", "Jabber Open Source Server: Client to
                 log_write(c2s->log, LOG_NOTICE, "attempting reconnect");
                 sleep(c2s->retry_sleep);
                 c2s_lost_router = 0;
+                if (c2s->router) sx_free(c2s->router);
                 _c2s_router_connect(c2s);
             }
 
@@ -757,6 +760,7 @@ JABBER_MAIN("jabberd2c2s", "Jabber 2 C2S", "Jabber Open Source Server: Client to
                 c2s->retry_left--;
                 sleep(c2s->retry_sleep);
                 c2s_lost_router = 0;
+                if (c2s->router) sx_free(c2s->router);
                 _c2s_router_connect(c2s);
             }
         }
@@ -767,6 +771,7 @@ JABBER_MAIN("jabberd2c2s", "Jabber 2 C2S", "Jabber Open Source Server: Client to
 
             /* free sess data */
             if(sess->ip != NULL) free(sess->ip);
+            if(sess->smcomp != NULL) free(sess->smcomp);
             if(sess->result != NULL) nad_free(sess->result);
             if(sess->resources != NULL)
                 for(res = sess->resources; res != NULL;) {
@@ -801,12 +806,15 @@ JABBER_MAIN("jabberd2c2s", "Jabber 2 C2S", "Jabber Open Source Server: Client to
 #endif
             if(c2s->packet_stats != NULL) {
                 int fd = open(c2s->packet_stats, O_TRUNC | O_CREAT | O_WRONLY, S_IRUSR | S_IWUSR | S_IRGRP);
-                if(fd) {
+                if (fd >= 0) {
                     char buf[100];
                     int len = snprintf(buf, 100, "%lld\n", c2s->packet_count);
-                    write(fd, buf, len);
-                    close(fd);
-                } else {
+                    if (write(fd, buf, len) != len) {
+                        close(fd);
+                        fd = -1;
+                    } else close(fd);
+                }
+                if (fd < 0) {
                     log_write(c2s->log, LOG_ERR, "failed to write packet statistics to: %s", c2s->packet_stats);
                     c2s_shutdown = 1;
                 }
@@ -821,9 +829,9 @@ JABBER_MAIN("jabberd2c2s", "Jabber 2 C2S", "Jabber Open Source Server: Client to
     if(xhash_iter_first(c2s->sessions))
         do {
             xhv.sess_val = &sess;
-            xhash_iter_get(c2s->sessions, NULL, xhv.val);
+            xhash_iter_get(c2s->sessions, NULL, NULL, xhv.val);
 
-            if(sess->active)
+            if(sess->active && sess->s)
                 sx_close(sess->s);
 
         } while(xhash_iter_next(c2s->sessions));
@@ -849,6 +857,7 @@ JABBER_MAIN("jabberd2c2s", "Jabber 2 C2S", "Jabber Open Source Server: Client to
     while(jqueue_size(c2s->dead) > 0)
         sx_free((sx_t) jqueue_pull(c2s->dead));
 
+    if (c2s->fd != NULL) mio_close(c2s->mio, c2s->fd);
     sx_free(c2s->router);
 
     sx_env_free(c2s->sx_env);
