@@ -64,7 +64,7 @@ int in_mio_callback(mio_t m, mio_action_t a, mio_fd_t fd, void *data, void *arg)
     conn_t in = (conn_t) arg;
     s2s_t s2s = (s2s_t) arg;
     struct sockaddr_storage sa;
-    int namelen = sizeof(sa), port, nbytes;
+    int namelen = sizeof(sa), port, nbytes, flags = 0;
     char ipport[INET6_ADDRSTRLEN + 17];
 
     switch(a) {
@@ -86,6 +86,8 @@ int in_mio_callback(mio_t m, mio_action_t a, mio_fd_t fd, void *data, void *arg)
         case action_CLOSE:
             log_debug(ZONE, "close action on fd %d", fd->fd);
 
+            if (fd == s2s->server_fd) break;
+
             /* !!! logging */
             log_write(in->s2s->log, LOG_NOTICE, "[%d] [%s, port=%d] disconnect, packets: %i", fd->fd, in->ip, in->port, in->packet_count);
 
@@ -138,11 +140,16 @@ int in_mio_callback(mio_t m, mio_action_t a, mio_fd_t fd, void *data, void *arg)
             snprintf(ipport, INET6_ADDRSTRLEN + 16, "%s/%d", in->ip, in->port);
             xhash_put(s2s->in_accept, pstrdup(xhash_pool(s2s->in_accept),ipport), (void *) in);
 
+            flags = S2S_DB_HEADER;
 #ifdef HAVE_SSL
-            sx_server_init(in->s, S2S_DB_HEADER | ((s2s->sx_ssl != NULL) ? SX_SSL_STARTTLS_OFFER : 0) );
-#else
-            sx_server_init(in->s, S2S_DB_HEADER);
+            if(s2s->sx_ssl != NULL)
+                flags |= SX_SSL_STARTTLS_OFFER;
 #endif
+#ifdef HAVE_LIBZ
+            if(s2s->compression)
+                flags |= SX_COMPRESS_OFFER;
+#endif
+            sx_server_init(in->s, flags);
             break;
     }
 
@@ -258,7 +265,7 @@ static int _in_sx_callback(sx_t s, sx_event_t e, void *data, void *arg) {
         case event_PACKET:
             /* we're counting packets */
             in->packet_count++;
-	    in->s2s->packet_count++;
+            in->s2s->packet_count++;
 
             nad = (nad_t) data;
 
@@ -498,7 +505,7 @@ static void _in_verify(conn_t in, nad_t nad) {
 
 /** they're trying to send us something */
 static void _in_packet(conn_t in, nad_t nad) {
-    int attr, ns, sns;
+    int elem, attr, ns, sns;
     jid_t from, to;
     char *rkey;
     
@@ -555,17 +562,20 @@ static void _in_packet(conn_t in, nad_t nad) {
      * handled by sm (and not just routed through the server), so set the
      * jabber:client namespace.
      */
-    ns = nad->elems[0].ns;
-    if(ns < 0) {
+    if(ns >= 0 || nad->elems[0].ns < 0) {
         ns = nad_add_namespace(nad, uri_CLIENT, NULL);
-        nad->scope = -1;
+        for(elem = 0; elem < nad->ecur; elem++)
+            if(nad->elems[elem].ns == ns)
+                nad->elems[elem].ns = nad->nss[nad->elems[elem].ns].next;
         nad->nss[ns].next = nad->elems[0].ns;
         nad->elems[0].ns = ns;
+        nad->scope = -1;
     }
-    nad->elems[0].my_ns = ns;
+
+    nad->elems[0].my_ns = nad->elems[0].ns;
 
     /* wrap up the packet */
-    ns = nad_add_namespace(nad, uri_COMPONENT, "comp");
+    ns = nad_add_namespace(nad, uri_COMPONENT, NULL);
 
     nad_wrap_elem(nad, 0, ns, "route");
 
