@@ -116,7 +116,7 @@ static void _s2s_config_expand(s2s_t s2s) {
     s2s->packet_stats = config_get_one(s2s->config, "stats.packet", 0);
 
     /*
-    * If no origin IP is specified, use local IP as the originating one:
+     * If no origin IP is specified, use local IP as the originating one:
      * it makes most sense, at least for SSL'ized connections.
      */
     s2s->local_ip = config_get_one(s2s->config, "local.ip", 0);
@@ -150,6 +150,8 @@ static void _s2s_config_expand(s2s_t s2s) {
 
     s2s->io_max_fds = j_atoi(config_get_one(s2s->config, "io.max_fds", 0), 1024);
 
+    s2s->compression = (config_get(s2s->config, "io.compression") != NULL);
+
     s2s->stanza_size_limit = j_atoi(config_get_one(s2s->config, "io.limits.stanzasize", 0), 0);
 
     s2s->check_interval = j_atoi(config_get_one(s2s->config, "check.interval", 0), 60);
@@ -171,6 +173,7 @@ static void _s2s_config_expand(s2s_t s2s) {
     if (s2s->dns_min_ttl < 5)
         s2s->dns_min_ttl = 5;
     s2s->dns_max_ttl = j_atoi(config_get_one(s2s->config, "lookup.max-ttl", 0), 86400);
+    s2s->etc_hosts_ttl = j_atoi(config_get_one(s2s->config, "lookup.etc-hosts-ttl", 0), 86400);
     s2s->out_reuse = config_count(s2s->config, "out-conn-reuse") ? 1 : 0;
 }
 
@@ -252,6 +255,7 @@ static int _s2s_router_connect(s2s_t s2s) {
 int _s2s_check_conn_routes(s2s_t s2s, conn_t conn, const char *direction)
 {
   char *rkey;
+  int rkeylen;
   conn_state_t state;
   time_t now, dialback_time;
 
@@ -261,17 +265,17 @@ int _s2s_check_conn_routes(s2s_t s2s, conn_t conn, const char *direction)
      do {
            /* retrieve state in a separate operation, as sizeof(int) != sizeof(void *) on 64-bit platforms,
               so passing a pointer to state in xhash_iter_get is unsafe */
-           xhash_iter_get(conn->states, (const char **) &rkey, NULL);
-           state = (conn_state_t) xhash_get(conn->states, rkey);
+           xhash_iter_get(conn->states, (const char **) &rkey, &rkeylen, NULL);
+           state = (conn_state_t) xhash_getx(conn->states, rkey, rkeylen);
 
            if (state == conn_INPROGRESS) {
-              dialback_time = (time_t) xhash_get(conn->states_time, rkey);
+              dialback_time = (time_t) xhash_getx(conn->states_time, rkey, rkeylen);
 
               if(now > dialback_time + s2s->check_queue) {
-                 log_write(s2s->log, LOG_NOTICE, "[%d] [%s, port=%d] dialback for %s route '%s' timed out", conn->fd->fd, conn->ip, conn->port, direction, rkey);
+                 log_write(s2s->log, LOG_NOTICE, "[%d] [%s, port=%d] dialback for %s route '%.*s' timed out", conn->fd->fd, conn->ip, conn->port, direction, rkeylen, rkey);
 
-                 xhash_zap(conn->states, rkey);
-                 xhash_zap(conn->states_time, rkey);
+                 xhash_zapx(conn->states, rkey, rkeylen);
+                 xhash_zapx(conn->states_time, rkey, rkeylen);
 
                  /* stream error */
                  sx_error(conn->s, stream_err_CONNECTION_TIMEOUT, "dialback timed out");
@@ -293,9 +297,11 @@ static void _s2s_time_checks(s2s_t s2s) {
     conn_t conn;
     time_t now;
     char *rkey, *key;
+    int keylen;
     jqueue_t q;
     dnscache_t dns;
     char *c;
+    int c_len;
     union xhashv xhv;
 
     now = time(NULL);
@@ -305,21 +311,22 @@ static void _s2s_time_checks(s2s_t s2s) {
         if(xhash_iter_first(s2s->outq))
             do {
                 xhv.jq_val = &q;
-                xhash_iter_get(s2s->outq, (const char **) &rkey, xhv.val);
+                xhash_iter_get(s2s->outq, (const char **) &rkey, &keylen, xhv.val);
 
-                log_debug(ZONE, "running time checks for %s", rkey);
-                c = strchr(rkey, '/');
+                log_debug(ZONE, "running time checks for %.*s", keylen, rkey);
+                c = memchr(rkey, '/', keylen);
                 c++;
+                c_len = keylen - (c - rkey);
 
                 /* dns lookup timeout check first */
-                dns = xhash_get(s2s->dnscache, c);
+                dns = xhash_getx(s2s->dnscache, c, c_len);
                 if(dns != NULL && dns->pending) {
-                    log_debug(ZONE, "dns lookup pending for %s", c);
+                    log_debug(ZONE, "dns lookup pending for %.*s", c_len, c);
                     if(now > dns->init_time + s2s->check_queue) {
-                        log_write(s2s->log, LOG_NOTICE, "dns lookup for %s timed out", c);
+                        log_write(s2s->log, LOG_NOTICE, "dns lookup for %.*s timed out", c_len, c);
 
                         /* bounce queue */
-                        out_bounce_route_queue(s2s, rkey, stanza_err_REMOTE_SERVER_NOT_FOUND);
+                        out_bounce_route_queue(s2s, rkey, keylen, stanza_err_REMOTE_SERVER_NOT_FOUND);
 
                         /* expire pending dns entry */
                         xhash_zap(s2s->dnscache, dns->name);
@@ -339,14 +346,14 @@ static void _s2s_time_checks(s2s_t s2s) {
                 }
 
                 /* get the conn */
-                conn = xhash_get(s2s->out_dest, c);
+                conn = xhash_getx(s2s->out_dest, c, c_len);
                 if(conn == NULL) {
                     if(jqueue_size(q) > 0) {
                        /* no pending conn? perhaps it failed? */
-                       log_debug(ZONE, "no pending connection for %s, bouncing %i packets in queue", c, jqueue_size(q));
+                       log_debug(ZONE, "no pending connection for %.*s, bouncing %i packets in queue", c_len, c, jqueue_size(q));
 
                        /* bounce queue */
-                       out_bounce_route_queue(s2s, rkey, stanza_err_REMOTE_SERVER_TIMEOUT);
+                       out_bounce_route_queue(s2s, rkey, keylen, stanza_err_REMOTE_SERVER_TIMEOUT);
                     }
 
                     continue;
@@ -388,10 +395,10 @@ static void _s2s_time_checks(s2s_t s2s) {
             if(xhash_iter_first(s2s->out_host))
                 do {
                     xhv.conn_val = &conn;
-                    xhash_iter_get(s2s->out_host, (const char **) &key, xhv.val);
-                    log_debug(ZONE, "checking dialback state for outgoing conn %s", key);
+                    xhash_iter_get(s2s->out_host, (const char **) &key, &keylen, xhv.val);
+                    log_debug(ZONE, "checking dialback state for outgoing conn %.*s", keylen, key);
                     if (_s2s_check_conn_routes(s2s, conn, "outgoing")) {
-                        log_debug(ZONE, "checking pending verify requests for outgoing conn %s", key);
+                        log_debug(ZONE, "checking pending verify requests for outgoing conn %.*s", keylen, key);
                         if (conn->verify > 0 && now > conn->last_verify + s2s->check_queue) {
                             log_write(s2s->log, LOG_NOTICE, "[%d] [%s, port=%d] dialback verify request timed out", conn->fd->fd, conn->ip, conn->port);
                             sx_error(conn->s, stream_err_CONNECTION_TIMEOUT, "dialback verify request timed out");
@@ -403,7 +410,7 @@ static void _s2s_time_checks(s2s_t s2s) {
             if(xhash_iter_first(s2s->out_dest))
                 do {
                     xhv.conn_val = &conn;
-                    xhash_iter_get(s2s->out_dest, (const char **) &key, xhv.val);
+                    xhash_iter_get(s2s->out_dest, (const char **) &key, &keylen, xhv.val);
                     log_debug(ZONE, "checking dialback state for outgoing conn %s (%s)", conn->dkey, conn->key);
                     if (_s2s_check_conn_routes(s2s, conn, "outgoing")) {
                         log_debug(ZONE, "checking pending verify requests for outgoing conn %s (%s)", conn->dkey, conn->key);
@@ -420,9 +427,9 @@ static void _s2s_time_checks(s2s_t s2s) {
         if(xhash_iter_first(s2s->in))
             do {
                 xhv.conn_val = &conn;
-                xhash_iter_get(s2s->in, (const char **) &key, xhv.val);
+                xhash_iter_get(s2s->in, (const char **) &key, &keylen, xhv.val);
 
-                log_debug(ZONE, "checking dialback state for incoming conn %s", key);
+                log_debug(ZONE, "checking dialback state for incoming conn %.*s", keylen, key);
                 if (_s2s_check_conn_routes(s2s, conn, "incoming"))
                     /* if the connection is still valid, check that dialbacks have been initiated */
                     if(!xhash_count(conn->states) && now > conn->init_time + s2s->check_queue) {
@@ -436,7 +443,7 @@ static void _s2s_time_checks(s2s_t s2s) {
         if(xhash_iter_first(s2s->in_accept))
             do {
                 xhv.conn_val = &conn;
-                xhash_iter_get(s2s->in_accept, (const char **) &key, xhv.val);
+                xhash_iter_get(s2s->in_accept, (const char **) &key, &keylen, xhv.val);
 
                 log_debug(ZONE, "checking stream connection state for incoming conn %i", conn->fd->fd);
                 if(!conn->online && now > conn->init_time + s2s->check_queue) {
@@ -452,7 +459,7 @@ static void _s2s_time_checks(s2s_t s2s) {
         if(xhash_iter_first(s2s->out_host))
             do {
                 xhv.conn_val = &conn;
-                xhash_iter_get(s2s->out_host, NULL, xhv.val);
+                xhash_iter_get(s2s->out_host, NULL, NULL, xhv.val);
 
                 if(s2s->check_keepalive > 0 && conn->last_activity > 0 && now > conn->last_activity + s2s->check_keepalive && conn->s->state >= state_STREAM) {
                     log_debug(ZONE, "sending keepalive for %d", conn->fd->fd);
@@ -464,7 +471,7 @@ static void _s2s_time_checks(s2s_t s2s) {
         if(xhash_iter_first(s2s->out_dest))
             do {
                 xhv.conn_val = &conn;
-                xhash_iter_get(s2s->out_dest, NULL, xhv.val);
+                xhash_iter_get(s2s->out_dest, NULL, NULL, xhv.val);
 
                 if(s2s->check_keepalive > 0 && conn->last_activity > 0 && now > conn->last_activity + s2s->check_keepalive && conn->s->state >= state_STREAM) {
                     log_debug(ZONE, "sending keepalive for %d", conn->fd->fd);
@@ -482,8 +489,8 @@ static void _s2s_time_checks(s2s_t s2s) {
             if(xhash_iter_first(s2s->out_host))
                 do {
                     xhv.conn_val = &conn;
-                    xhash_iter_get(s2s->out_host, (const char **) &key, xhv.val);
-                    log_debug(ZONE, "checking idle state for %s", key);
+                    xhash_iter_get(s2s->out_host, (const char **) &key, &keylen, xhv.val);
+                    log_debug(ZONE, "checking idle state for %.*s", keylen, key);
                     if (conn->last_packet > 0 && now > conn->last_packet + s2s->check_idle && conn->s->state >= state_STREAM) {
                         log_write(s2s->log, LOG_NOTICE, "[%d] [%s, port=%d] idle timeout", conn->fd->fd, conn->ip, conn->port);
                         sx_close(conn->s);
@@ -493,7 +500,7 @@ static void _s2s_time_checks(s2s_t s2s) {
             if(xhash_iter_first(s2s->out_dest))
                 do {
                     xhv.conn_val = &conn;
-                    xhash_iter_get(s2s->out_dest, (const char **) &key, xhv.val);
+                    xhash_iter_get(s2s->out_dest, (const char **) &key, &keylen, xhv.val);
                     log_debug(ZONE, "checking idle state for %s (%s)", conn->dkey, conn->key);
                     if (conn->last_packet > 0 && now > conn->last_packet + s2s->check_idle && conn->s->state >= state_STREAM) {
                         log_write(s2s->log, LOG_NOTICE, "[%d] [%s, port=%d] idle timeout", conn->fd->fd, conn->ip, conn->port);
@@ -506,8 +513,8 @@ static void _s2s_time_checks(s2s_t s2s) {
         if(xhash_iter_first(s2s->in))
             do {
                 xhv.conn_val = &conn;
-                xhash_iter_get(s2s->in, (const char **) &key, xhv.val);
-                log_debug(ZONE, "checking idle state for %s", key);
+                xhash_iter_get(s2s->in, (const char **) &key, &keylen, xhv.val);
+                log_debug(ZONE, "checking idle state for %.*s", keylen, key);
                 if (conn->last_packet > 0 && now > conn->last_packet + s2s->check_idle && conn->s->state >= state_STREAM) {
                     log_write(s2s->log, LOG_NOTICE, "[%d] [%s, port=%d] idle timeout", conn->fd->fd, conn->ip, conn->port);
                     sx_close(conn->s);
@@ -521,9 +528,8 @@ static void _s2s_time_checks(s2s_t s2s) {
 
 static void _s2s_dns_expiry(s2s_t s2s) {
     time_t now;
-    char *key;
-    dnscache_t dns;
-    dnsres_t res;
+    dnscache_t dns = NULL;
+    dnsres_t res = NULL;
     union xhashv xhv;
 
     now = time(NULL);
@@ -532,8 +538,8 @@ static void _s2s_dns_expiry(s2s_t s2s) {
     if(xhash_iter_first(s2s->dnscache))
         do {
             xhv.dns_val = &dns;
-            xhash_iter_get(s2s->dnscache, (const char **) &key, xhv.val);
-            if (!dns->pending && now > dns->expiry) {
+            xhash_iter_get(s2s->dnscache, NULL, NULL, xhv.val);
+            if (dns && !dns->pending && now > dns->expiry) {
                 log_debug(ZONE, "expiring DNS cache for %s", dns->name);
                 xhash_iter_zap(s2s->dnscache);
 
@@ -548,19 +554,25 @@ static void _s2s_dns_expiry(s2s_t s2s) {
                 }
                 free(dns);
             }
+            else if (dns == NULL) {
+                xhash_iter_zap(s2s->dnscache);
+            }
         } while(xhash_iter_next(s2s->dnscache));
 
     if(xhash_iter_first(s2s->dns_bad))
         do {
             xhv.dnsres_val = &res;
-            xhash_iter_get(s2s->dns_bad, (const char **) &key, xhv.val);
-            if (now > res->expiry) {
+            xhash_iter_get(s2s->dns_bad, NULL, NULL, xhv.val);
+            if (res && now > res->expiry) {
                 log_debug(ZONE, "expiring DNS bad host %s", res->key);
                 xhash_iter_zap(s2s->dns_bad);
 
                 free(res->key);
                 free(res);
             }
+            else if (res == NULL) {
+                xhash_iter_zap(s2s->dns_bad);
+            }
         } while(xhash_iter_next(s2s->dns_bad));
 }
 /** responses from the resolver */
@@ -599,19 +611,19 @@ JABBER_MAIN("jabberd2s2s", "Jabber 2 S2S", "Jabber Open Source Server: Server to
 
 #ifdef HAVE_WINSOCK2_H
 /* get winsock running */
-	{
-		WORD wVersionRequested;
-		WSADATA wsaData;
-		int err;
-		
-		wVersionRequested = MAKEWORD( 2, 2 );
-		
-		err = WSAStartup( wVersionRequested, &wsaData );
-		if ( err != 0 ) {
+    {
+        WORD wVersionRequested;
+        WSADATA wsaData;
+        int err;
+        
+        wVersionRequested = MAKEWORD( 2, 2 );
+        
+        err = WSAStartup( wVersionRequested, &wsaData );
+        if ( err != 0 ) {
             /* !!! tell user that we couldn't find a usable winsock dll */
-			return 0;
-		}
-	}
+            return 0;
+        }
+    }
 #endif
 
     jabber_signal(SIGINT, _s2s_signal);
@@ -711,6 +723,12 @@ JABBER_MAIN("jabberd2s2s", "Jabber 2 S2S", "Jabber Open Source Server: Server to
     }
 #endif
 
+#ifdef HAVE_LIBZ
+    /* get compression up and running */
+    if(s2s->compression)
+        sx_env_plugin(s2s->sx_env, sx_compress_init);
+#endif
+
     /* get sasl online */
     s2s->sx_sasl = sx_env_plugin(s2s->sx_env, sx_sasl_init, "xmpp", NULL, NULL);
     if(s2s->sx_sasl == NULL) {
@@ -754,6 +772,7 @@ JABBER_MAIN("jabberd2s2s", "Jabber 2 S2S", "Jabber Open Source Server: Server to
                 log_write(s2s->log, LOG_NOTICE, "attempting reconnect");
                 sleep(s2s->retry_sleep);
                 s2s_lost_router = 0;
+                if (s2s->router) sx_free(s2s->router);
                 _s2s_router_connect(s2s);
             }
 
@@ -766,12 +785,13 @@ JABBER_MAIN("jabberd2s2s", "Jabber 2 S2S", "Jabber Open Source Server: Server to
                 s2s->retry_left--;
                 sleep(s2s->retry_sleep);
                 s2s_lost_router = 0;
+                if (s2s->router) sx_free(s2s->router);
                 _s2s_router_connect(s2s);
             }
         }
 
         /* this has to be read unconditionally - we could receive replies to queries we cancelled */
-      	mio_read(s2s->mio, s2s->udns_mio_fd);
+          mio_read(s2s->mio, s2s->udns_mio_fd);
             
         /* cleanup dead sx_ts */
         while(jqueue_size(s2s->dead) > 0)
@@ -837,31 +857,43 @@ JABBER_MAIN("jabberd2s2s", "Jabber 2 S2S", "Jabber Open Source Server: Server to
     if(s2s->out_reuse) {
         if(xhash_iter_first(s2s->out_host))
             do {
-                xhash_iter_get(s2s->out_host, NULL, xhv.val);
-                sx_error(conn->s, stream_err_SYSTEM_SHUTDOWN, "s2s shutdown");
-                sx_close(conn->s);
-            } while(xhash_count(s2s->out_host));
+                xhash_iter_get(s2s->out_host, NULL, NULL, xhv.val);
+                if(conn) {
+                    sx_error(conn->s, stream_err_SYSTEM_SHUTDOWN, "s2s shutdown");
+                    out_bounce_conn_queues(conn, stanza_err_SERVICE_UNAVAILABLE);
+                    sx_close(conn->s);
+                }
+            } while(xhash_iter_next(s2s->out_host));
     } else {
         if(xhash_iter_first(s2s->out_dest))
             do {
-                xhash_iter_get(s2s->out_dest, NULL, xhv.val);
-                sx_error(conn->s, stream_err_SYSTEM_SHUTDOWN, "s2s shutdown");
-                sx_close(conn->s);
-            } while(xhash_count(s2s->out_dest));
+                xhash_iter_get(s2s->out_dest, NULL, NULL, xhv.val);
+                if(conn) {
+                    sx_error(conn->s, stream_err_SYSTEM_SHUTDOWN, "s2s shutdown");
+                    out_bounce_conn_queues(conn, stanza_err_SERVICE_UNAVAILABLE);
+                    sx_close(conn->s);
+                }
+            } while(xhash_iter_next(s2s->out_dest));
     }
 
     if(xhash_iter_first(s2s->in))
         do {
-            xhash_iter_get(s2s->in, NULL, xhv.val);
-            sx_error(conn->s, stream_err_SYSTEM_SHUTDOWN, "s2s shutdown");
-            sx_close(conn->s);
-        } while(xhash_count(s2s->in));
+            xhash_iter_get(s2s->in, NULL, NULL, xhv.val);
+            if(conn) {
+                sx_error(conn->s, stream_err_SYSTEM_SHUTDOWN, "s2s shutdown");
+                out_bounce_conn_queues(conn, stanza_err_SERVICE_UNAVAILABLE);
+                sx_close(conn->s);
+            }
+        } while(xhash_iter_next(s2s->in));
 
     if(xhash_iter_first(s2s->in_accept))
         do {
-            xhash_iter_get(s2s->in_accept, NULL, xhv.val);
-            sx_close(conn->s);
-        } while(xhash_count(s2s->in_accept));
+            xhash_iter_get(s2s->in_accept, NULL, NULL, xhv.val);
+            if(conn) {
+                out_bounce_conn_queues(conn, stanza_err_SERVICE_UNAVAILABLE);
+                sx_close(conn->s);
+            }
+        } while(xhash_iter_next(s2s->in_accept));
 
 
     /* remove dead streams */
@@ -876,6 +908,7 @@ JABBER_MAIN("jabberd2s2s", "Jabber 2 S2S", "Jabber Open Source Server: Server to
         xhash_free(conn->routes);
 
         if(conn->key != NULL) free(conn->key);
+        if(conn->dkey != NULL) free(conn->dkey);
         free(conn);
     }
 
@@ -883,7 +916,9 @@ JABBER_MAIN("jabberd2s2s", "Jabber 2 S2S", "Jabber Open Source Server: Server to
     xhv.jq_val = &q;
     if(xhash_iter_first(s2s->outq))
         do {
-             xhash_iter_get(s2s->outq, NULL, xhv.val);
+             xhash_iter_get(s2s->outq, NULL, NULL, xhv.val);
+             while (jqueue_size(q) > 0)
+                 out_pkt_free((pkt_t) jqueue_pull(q));
              free(q->key);
              jqueue_free(q);
         } while(xhash_iter_next(s2s->outq));
@@ -892,7 +927,7 @@ JABBER_MAIN("jabberd2s2s", "Jabber 2 S2S", "Jabber Open Source Server: Server to
     xhv.dns_val = &dns;
     if(xhash_iter_first(s2s->dnscache))
         do {
-             xhash_iter_get(s2s->dnscache, NULL, xhv.val);
+             xhash_iter_get(s2s->dnscache, NULL, NULL, xhv.val);
              xhash_free(dns->results);
              if (dns->query != NULL) {
                  if (dns->query->query != NULL)
@@ -908,7 +943,7 @@ JABBER_MAIN("jabberd2s2s", "Jabber 2 S2S", "Jabber Open Source Server: Server to
     xhv.dnsres_val = &res;
     if(xhash_iter_first(s2s->dns_bad))
         do {
-             xhash_iter_get(s2s->dns_bad, NULL, xhv.val);
+             xhash_iter_get(s2s->dns_bad, NULL, NULL, xhv.val);
              free(res->key);
              free(res);
         } while(xhash_iter_next(s2s->dns_bad));
@@ -918,8 +953,11 @@ JABBER_MAIN("jabberd2s2s", "Jabber 2 S2S", "Jabber Open Source Server: Server to
     dns_close(NULL);
 
     /* close mio */
+    mio_close(s2s->mio, s2s->udns_mio_fd);
     if(s2s->fd != NULL)
         mio_close(s2s->mio, s2s->fd);
+    if(s2s->server_fd != NULL)
+        mio_close(s2s->mio, s2s->server_fd);
 
     /* free hashes */
     xhash_free(s2s->outq);
