@@ -18,6 +18,9 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA02111-1307USA
  */
 
+#define _GNU_SOURCE
+#include <string.h>
+
 #include "s2s.h"
 
 #include <idna.h>
@@ -124,35 +127,36 @@ static void _out_packet_queue(s2s_t s2s, pkt_t pkt) {
     jqueue_push(q, (void *) pkt, 0);
 }
 
-static void _out_dialback(conn_t out, char *rkey) {
-    char *c, *dbkey;
+static void _out_dialback(conn_t out, char *rkey, int rkeylen) {
+    char *c, *dbkey, *tmp;
     nad_t nad;
-    int ns;
+    int elem, ns;
+    int from_len, to_len;
     time_t now;
 
     now = time(NULL);
 
-    c = strchr(rkey, '/');
-    *c = '\0';
+    c = memchr(rkey, '/', rkeylen);
+    from_len = c - rkey;
     c++;
+    to_len = rkeylen - (c - rkey);
     
     /* kick off the dialback */
-    dbkey = s2s_db_key(NULL, out->s2s->local_secret, c, out->s->id);
+    tmp = strndup(c, to_len);
+    dbkey = s2s_db_key(NULL, out->s2s->local_secret, tmp, out->s->id);
+    free(tmp);
 
     nad = nad_new();
 
     /* request auth */
     ns = nad_add_namespace(nad, uri_DIALBACK, "db");
-    nad_append_elem(nad, ns, "result", 0);
-    nad_append_attr(nad, -1, "from", rkey);
-    nad_append_attr(nad, -1, "to", c);
+    elem = nad_append_elem(nad, ns, "result", 0);
+    nad_set_attr(nad, elem, -1, "from", rkey, from_len);
+    nad_set_attr(nad, elem, -1, "to", c, to_len);
     nad_append_cdata(nad, dbkey, strlen(dbkey), 1);
 
-    c--;
-    *c = '/';
-
-    log_debug(ZONE, "sending auth request for %s (key %s)", rkey, dbkey);
-    log_write(out->s2s->log, LOG_NOTICE, "[%d] [%s, port=%d] sending dialback auth request for route '%s'", out->fd->fd, out->ip, out->port, rkey);
+    log_debug(ZONE, "sending auth request for %.*s (key %s)", rkeylen, rkey, dbkey);
+    log_write(out->s2s->log, LOG_NOTICE, "[%d] [%s, port=%d] sending dialback auth request for route '%.*s'", out->fd->fd, out->ip, out->port, rkeylen, rkey);
 
     /* off it goes */
     sx_nad_write(out->s, nad);
@@ -160,10 +164,10 @@ static void _out_dialback(conn_t out, char *rkey) {
     free(dbkey);
             
     /* we're in progress now */
-    xhash_put(out->states, pstrdup(xhash_pool(out->states), rkey), (void *) conn_INPROGRESS);
+    xhash_put(out->states, pstrdupx(xhash_pool(out->states), rkey, rkeylen), (void *) conn_INPROGRESS);
 
     /* record the time that we set conn_INPROGRESS state */
-    xhash_put(out->states_time, pstrdup(xhash_pool(out->states_time), rkey), (void *) now);
+    xhash_put(out->states_time, pstrdupx(xhash_pool(out->states_time), rkey, rkeylen), (void *) now);
 }
 
 int dns_select(s2s_t s2s, char *ip, int *port, time_t now, dnscache_t dns, int allow_bad) {
@@ -177,13 +181,16 @@ int dns_select(s2s_t s2s, char *ip, int *port, time_t now, dnscache_t dns, int a
     int rw_aaaa[DNS_MAX_RESULTS];
     int rw_a[DNS_MAX_RESULTS];
     int s_reuse = 0, s_aaaa = 0, s_a = 0, s_bad = 0; /* count */
-    int p_reuse, p_aaaa, p_a; /* list prio */
-    int wt_reuse, wt_aaaa, wt_a; /* weight total */
+    int p_reuse = 0, p_aaaa = 0, p_a = 0; /* list prio */
+    int wt_reuse = 0, wt_aaaa = 0, wt_a = 0; /* weight total */
     int c_expired_good = 0;
     union xhashv xhv;
     dnsres_t res;
     char *ipport;
+    int ipport_len;
     char *c;
+    int c_len;
+    char *tmp;
 
     /* for all results:
      * - if not expired
@@ -204,10 +211,10 @@ int dns_select(s2s_t s2s, char *ip, int *port, time_t now, dnscache_t dns, int a
     if (xhash_iter_first(dns->results)) {
         dnsres_t bad = NULL;
         do {
-            xhash_iter_get(dns->results, (const char **) &ipport, xhv.val);
+            xhash_iter_get(dns->results, (const char **) &ipport, &ipport_len, xhv.val);
 
             if (s2s->dns_bad_timeout > 0)
-                bad = xhash_get(s2s->dns_bad, ipport);
+                bad = xhash_getx(s2s->dns_bad, ipport, ipport_len);
 
             if (now > res->expiry) {
                 /* good host? */
@@ -221,7 +228,7 @@ int dns_select(s2s_t s2s, char *ip, int *port, time_t now, dnscache_t dns, int a
                 l_bad[s_bad++] = res;
 
                 log_debug(ZONE, "host '%s' bad", res->key);
-            } else if (s2s->out_reuse && xhash_get(s2s->out_host, ipport) != NULL) {
+            } else if (s2s->out_reuse && xhash_getx(s2s->out_host, ipport, ipport_len) != NULL) {
                 /* existing connection */
                 log_debug(ZONE, "host '%s' exists", res->key);
                 if (s_reuse == 0 || p_reuse > res->prio) {
@@ -242,7 +249,7 @@ int dns_select(s2s_t s2s, char *ip, int *port, time_t now, dnscache_t dns, int a
                 } else {
                     log_debug(ZONE, "ignored host with prio %d", res->prio);
                 }
-            } else if (strchr(ipport, ':') != NULL) {
+            } else if (memchr(ipport, ':', ipport_len) != NULL) {
                 /* ipv6 */
                 log_debug(ZONE, "host '%s' IPv6", res->key);
                 if (s_aaaa == 0 || p_aaaa > res->prio) {
@@ -365,30 +372,33 @@ int dns_select(s2s_t s2s, char *ip, int *port, time_t now, dnscache_t dns, int a
     assert(ipport != NULL);
 
     /* copy the ip and port to the packet */
+    ipport_len = strlen(ipport);
     c = strchr(ipport, '/');
-    *c = '\0';
+    strncpy(ip, ipport, c-ipport);
+    ip[c-ipport] = '\0';
     c++;
-
-    strcpy(ip, ipport);
-    *port = atoi(c);
-
-    c--;
-    *c = '/';
+    c_len = ipport_len - (c - ipport);
+    tmp = strndup(c, c_len);
+    *port = atoi(tmp);
+    free(tmp);
 
     return 0;
 }
 
 /** find/make a connection for a route */
-int out_route(s2s_t s2s, char *route, conn_t *out, int allow_bad) {
+int out_route(s2s_t s2s, char *route, int routelen, conn_t *out, int allow_bad) {
     dnscache_t dns;
     char ipport[INET6_ADDRSTRLEN + 16], *dkey, *c;
     time_t now;
     int reuse = 0;
     char ip[INET6_ADDRSTRLEN];
-    int port;
+    int port, c_len, from_len;
 
-    c = strchr(route, '/');
-    dkey = strdup(c+1);
+    c = memchr(route, '/', routelen);
+    from_len = c - route;
+    c++;
+    c_len = routelen - (c - route);
+    dkey = strndup(c, c_len);
 
     log_debug(ZONE, "trying to find connection for '%s'", dkey);
     *out = (conn_t) xhash_get(s2s->out_dest, dkey);
@@ -498,7 +508,7 @@ int out_route(s2s_t s2s, char *route, conn_t *out, int allow_bad) {
                 xhash_put(s2s->out_host, (*out)->key, (void *) *out);
             xhash_put(s2s->out_dest, s2s->out_reuse ? pstrdup(xhash_pool((*out)->routes), dkey) : dkey, (void *) *out);
 
-            xhash_put((*out)->routes, pstrdup(xhash_pool((*out)->routes), route), (void *) 1);
+            xhash_put((*out)->routes, pstrdupx(xhash_pool((*out)->routes), route, routelen), (void *) 1);
 
             /* connect */
             log_debug(ZONE, "initiating connection to %s", ipport);
@@ -536,7 +546,7 @@ int out_route(s2s_t s2s, char *route, conn_t *out, int allow_bad) {
                 *out = NULL;
 
                 /* try again without allowing bad hosts */
-                return out_route(s2s, route, out, 0);
+                return out_route(s2s, route, routelen, out, 0);
             } else {
                 log_write(s2s->log, LOG_NOTICE, "[%d] [%s, port=%d] outgoing connection for '%s'", (*out)->fd->fd, (*out)->ip, (*out)->port, dkey);
 
@@ -545,9 +555,7 @@ int out_route(s2s_t s2s, char *route, conn_t *out, int allow_bad) {
 #ifdef HAVE_SSL
                 /* Send a stream version of 1.0 if we can do STARTTLS */
                 if(s2s->sx_ssl != NULL) {
-                    *c = '\0';
-                    sx_client_init((*out)->s, S2S_DB_HEADER, uri_SERVER, dkey, route, "1.0");
-                    *c = '/';
+                    sx_client_init((*out)->s, S2S_DB_HEADER, uri_SERVER, dkey, pstrdupx(xhash_pool((*out)->routes), route, from_len), "1.0");
                 } else {
                     sx_client_init((*out)->s, S2S_DB_HEADER, uri_SERVER, NULL, NULL, NULL);
                 }
@@ -564,26 +572,36 @@ int out_route(s2s_t s2s, char *route, conn_t *out, int allow_bad) {
 
     /* connection in progress, or re-using connection: add to routes list */
     if (!(*out)->online || reuse) {
-        if (xhash_get((*out)->routes, route) == NULL)
-            xhash_put((*out)->routes, pstrdup(xhash_pool((*out)->routes), route), (void *) 1);
+        if (xhash_getx((*out)->routes, route, routelen) == NULL)
+            xhash_put((*out)->routes, pstrdupx(xhash_pool((*out)->routes), route, routelen), (void *) 1);
     }
 
     free(dkey);
     return 0;
 }
 
+void out_pkt_free(pkt_t pkt)
+{
+    nad_free(pkt->nad);
+    jid_free(pkt->from);
+    jid_free(pkt->to);
+    free(pkt);
+}
+
 /** send a packet out */
 int out_packet(s2s_t s2s, pkt_t pkt) {
     char *rkey;
+    int rkeylen;
     conn_t out;
     conn_state_t state;
     int ret;
 
     /* new route key */
     rkey = s2s_route_key(NULL, pkt->from->domain, pkt->to->domain);
+    rkeylen = strlen(rkey);
 
     /* get a connection */
-    ret = out_route(s2s, rkey, &out, 1);
+    ret = out_route(s2s, rkey, rkeylen, &out, 1);
 
     if (out == NULL) {
         /* connection not available, queue packet */
@@ -592,7 +610,7 @@ int out_packet(s2s_t s2s, pkt_t pkt) {
         /* check if out_route was successful in attempting a connection */
         if (ret) {
             /* bounce queue */
-            out_bounce_route_queue(s2s, rkey, stanza_err_SERVICE_UNAVAILABLE);
+            out_bounce_route_queue(s2s, rkey, rkeylen, stanza_err_SERVICE_UNAVAILABLE);
 
             free(rkey);
             return -1;
@@ -665,7 +683,7 @@ int out_packet(s2s_t s2s, pkt_t pkt) {
     }
 
     /* this is a new route - send dialback auth request to piggyback on the existing connection */
-    _out_dialback(out, rkey);
+    _out_dialback(out, rkey, rkeylen);
 
     free(rkey);
     return 0;
@@ -909,6 +927,82 @@ static void _dns_result_aaaa(struct dns_ctx *ctx, struct dns_rr_a6 *result, void
     free(result);
 }
 
+/* try /etc/hosts if the A process did not return any results */
+static int _etc_hosts_lookup(const char *cszName, char *szIP, const int ciMaxIPLen) {
+#define EHL_LINE_LEN 260
+    int iSuccess = 0;
+    size_t iLen;
+    char szLine[EHL_LINE_LEN + 1]; /* one extra for the space character (*) */
+    char *pcStart, *pcEnd;
+    FILE *fHosts;
+
+    do {
+        /* initialization */
+        fHosts = NULL;
+
+        /* sanity checks */
+        if ((cszName == NULL) || (szIP == NULL) || (ciMaxIPLen <= 0))
+            break;
+        szIP[0] = 0;
+
+        /* open the hosts file */
+#ifdef _WIN32
+        pcStart = getenv("WINDIR");
+        if (pcStart != NULL) {
+            sprintf(szLine, "%s\\system32\\drivers\\etc\\hosts", pcStart);
+        } else {
+            strcpy(szLine, "C:\\WINDOWS\\system32\\drivers\\etc\\hosts");
+        }
+#else
+        strcpy(szLine, "/etc/hosts");
+#endif
+        fHosts = fopen(szLine, "r");
+        if (fHosts == NULL)
+            break;
+
+        /* read line by line ... */
+        while (fgets(szLine, EHL_LINE_LEN, fHosts) != NULL) {
+            /* remove comments */
+            pcStart = strchr (szLine, '#');
+            if (pcStart != NULL)
+                *pcStart = 0;
+            strcat(szLine, " "); /* append a space character for easier parsing (*) */
+
+            /* first to appear: IP address */
+            iLen = strspn(szLine, "1234567890.");
+            if ((iLen < 7) || (iLen > 15)) /* superficial test for anything between x.x.x.x and xxx.xxx.xxx.xxx */
+                continue;
+            pcEnd = szLine + iLen;
+            *pcEnd = 0;
+            pcEnd++; /* not beyond the end of the line yet (*) */
+
+            /* check strings separated by blanks, tabs or newlines */
+            pcStart = pcEnd + strspn(pcEnd, " \t\n");
+            while (*pcStart != 0) {
+                pcEnd = pcStart + strcspn(pcStart, " \t\n");
+                *pcEnd = 0;
+                pcEnd++; /* not beyond the end of the line yet (*) */
+
+                if (strcasecmp(pcStart, cszName) == 0) {
+                    strncpy(szIP, szLine, ciMaxIPLen - 1);
+                    szIP[ciMaxIPLen - 1] = '\0';
+                    iSuccess = 1;
+                    break;
+                }
+
+                pcStart = pcEnd + strspn(pcEnd, " \t\n");
+            }
+            if (iSuccess)
+                break;
+        }
+    } while (0);
+
+    if (fHosts != NULL)
+        fclose(fHosts);
+
+    return (iSuccess);
+}
+
 /* this function is called with a NULL ctx to start the A/AAAA process */
 static void _dns_result_a(struct dns_ctx *ctx, struct dns_rr_a4 *result, void *data) {
     dnsquery_t query = data;
@@ -916,8 +1010,18 @@ static void _dns_result_a(struct dns_ctx *ctx, struct dns_rr_a4 *result, void *d
     query->query = NULL;
 
     if (ctx != NULL && result == NULL) {
-        log_debug(ZONE, "dns failure for %s@%p: A %s (%d)", query->name, query,
-            query->cur_host, dns_status(ctx));
+#define DRA_IP_LEN 16
+        char szIP[DRA_IP_LEN];
+        if (_etc_hosts_lookup (query->name, szIP, DRA_IP_LEN)) {
+            log_debug(ZONE, "/etc/lookup for %s@%p: %s (%d)", query->name,
+                query, szIP, query->s2s->etc_hosts_ttl);
+
+            _dns_add_result (query, szIP, query->cur_port,
+                query->cur_prio, query->cur_weight, query->s2s->etc_hosts_ttl);
+        } else {
+            log_debug(ZONE, "dns failure for %s@%p: A %s (%d)", query->name, query,
+                query->cur_host, dns_status(ctx));
+        }
     } else if (result != NULL) {
         char ip[INET_ADDRSTRLEN];
         int i;
@@ -943,57 +1047,57 @@ static void _dns_result_a(struct dns_ctx *ctx, struct dns_rr_a4 *result, void *d
 
     /* resolve the next host in the list */
     if (xhash_iter_first(query->hosts)) {
-        char *ipport, *c;
+        char *ipport, *c, *tmp;
+        int ipport_len, ip_len, port_len;
         dnsres_t res;
         union xhashv xhv;
 
         xhv.dnsres_val = &res;
 
         /* get the first entry */
-        xhash_iter_get(query->hosts, (const char **) &ipport, xhv.val);
+        xhash_iter_get(query->hosts, (const char **) &ipport, &ipport_len, xhv.val);
 
         /* remove the host from the list */
         xhash_iter_zap(query->hosts);
 
-        c = strchr(ipport, '/');
-        *c = '\0';
+        c = memchr(ipport, '/', ipport_len);
+        ip_len = c - ipport;
         c++;
+        port_len = ipport_len - (c - ipport);
 
         /* resolve hostname */
         free(query->cur_host);
-        query->cur_host = strdup(ipport);
-        query->cur_port = atoi(c);
+        query->cur_host = strndup(ipport, ip_len);
+        tmp = strndup(c, port_len);
+        query->cur_port = atoi(tmp);
+        free(tmp);
         query->cur_prio = res->prio;
         query->cur_weight = res->weight;
         query->cur_expiry = res->expiry;
         log_debug(ZONE, "dns ttl for %s@%p limited to %d", query->name, query, query->cur_expiry);
 
         if (query->s2s->resolve_aaaa) {
-            log_debug(ZONE, "dns request for %s@%p: AAAA %s", query->name, query, ipport);
+            log_debug(ZONE, "dns request for %s@%p: AAAA %s", query->name, query, query->cur_host);
 
-            query->query = dns_submit_a6(NULL, ipport,
-                DNS_NOSRCH, _dns_result_aaaa, query);
+            query->query = dns_submit_a6(NULL, query->cur_host, DNS_NOSRCH, _dns_result_aaaa, query);
 
             /* if submit failed, call ourselves with a NULL result */
             if (query->query == NULL)
                 _dns_result_aaaa(ctx, NULL, query);
         } else {
-            log_debug(ZONE, "dns request for %s@%p: A %s", query->name, query, ipport);
+            log_debug(ZONE, "dns request for %s@%p: A %s", query->name, query, query->cur_host);
 
-            query->query = dns_submit_a4(NULL, ipport,
-                DNS_NOSRCH, _dns_result_a, query);
+            query->query = dns_submit_a4(NULL, query->cur_host, DNS_NOSRCH, _dns_result_a, query);
 
             /* if submit failed, call ourselves with a NULL result */
             if (query->query == NULL)
                 _dns_result_a(ctx, NULL, query);
         }
 
-        *c = '/';
-        c--;
-
     /* finished */
     } else {
         time_t now = time(NULL);
+        char *domain;
 
         free(query->cur_host);
         query->cur_host = NULL;
@@ -1018,7 +1122,7 @@ static void _dns_result_a(struct dns_ctx *ctx, struct dns_rr_a4 *result, void *d
             xhv.dnsres_val = &res;
 
             do {
-                xhash_iter_get(query->results, NULL, xhv.val);
+                xhash_iter_get(query->results, NULL, NULL, xhv.val);
 
                 if (res->expiry > query->s2s->dns_max_ttl)
                     res->expiry = query->s2s->dns_max_ttl;
@@ -1032,11 +1136,13 @@ static void _dns_result_a(struct dns_ctx *ctx, struct dns_rr_a4 *result, void *d
 
         xhash_free(query->hosts);
         query->hosts = NULL;
-        char *domain;
         if (idna_to_unicode_8z8z(query->name, &domain, 0) != IDNA_SUCCESS)
         {
             log_write(query->s2s->log, LOG_ERR, "idna dns decode for %s failed", query->name);
-            /* TODO: Is it better to shortcut resolution failure here? */
+            /* fake empty results to shortcut resolution failure */
+            xhash_free(query->results);
+            query->results = xhash_new(71);
+            query->expiry = time(NULL) + 99999999;
             domain = strdup(query->name);
         }
         out_resolve(query->s2s, domain, query->results, query->expiry);
@@ -1050,15 +1156,17 @@ void dns_resolve_domain(s2s_t s2s, dnscache_t dns) {
     dnsquery_t query = (dnsquery_t) calloc(1, sizeof(struct dnsquery_st));
 
     query->s2s = s2s;
+    query->results = xhash_new(71);
     if (idna_to_ascii_8z(dns->name, &query->name, 0) != IDNA_SUCCESS)
     {
         log_write(s2s->log, LOG_ERR, "idna dns encode for %s failed", dns->name);
-        /* TODO: Is it better to shortcut resolution failure here? */
-        query->name = strdup(dns->name);
+        /* shortcut resolution failure */
+        query->expiry = time(NULL) + 99999999;
+        out_resolve(query->s2s, dns->name, query->results, query->expiry);
+        return;
     }
-    query->srv_i = -1;
     query->hosts = xhash_new(71);
-    query->results = xhash_new(71);
+    query->srv_i = -1;
     query->expiry = 0;
     query->cur_host = NULL;
     query->cur_port = 0;
@@ -1176,57 +1284,61 @@ static int _out_mio_callback(mio_t m, mio_action_t a, mio_fd_t fd, void *data, v
 
             if (xhash_iter_first(out->routes)) {
                 char *rkey;
+                int rkeylen;
                 char *c;
+                int c_len;
 
                 /* remove all the out_dest entries */
                 do {
-                    xhash_iter_get(out->routes, (const char **) &rkey, NULL);
-                    c = strchr(rkey, '/');
+                    xhash_iter_get(out->routes, (const char **) &rkey, &rkeylen, NULL);
+                    c = memchr(rkey, '/', rkeylen);
                     c++;
+                    c_len = rkeylen - (c - rkey);
 
-                    log_debug(ZONE, "route '%s'", rkey);
-                    if (xhash_get(out->s2s->out_dest, c) != NULL) {
-                        log_debug(ZONE, "removing dest entry for '%s'", c);
-                        xhash_zap(out->s2s->out_dest, c);
+                    log_debug(ZONE, "route '%.*s'", rkeylen, rkey);
+                    if (xhash_getx(out->s2s->out_dest, c, c_len) != NULL) {
+                        log_debug(ZONE, "removing dest entry for '%.*s'", c_len, c);
+                        xhash_zapx(out->s2s->out_dest, c, c_len);
                     }
                 } while(xhash_iter_next(out->routes));
             }
 
             if (xhash_iter_first(out->routes)) {
                 char *rkey;
+                int rkeylen;
                 jqueue_t q;
                 int npkt;
 
                 /* retry all the routes */
                 do {
-                    xhash_iter_get(out->routes, (const char **) &rkey, NULL);
+                    xhash_iter_get(out->routes, (const char **) &rkey, &rkeylen, NULL);
 
-                    q = xhash_get(out->s2s->outq, rkey);
+                    q = xhash_getx(out->s2s->outq, rkey, rkeylen);
                     if (out->s2s->retry_limit > 0 && q != NULL && jqueue_age(q) > out->s2s->retry_limit) {
-                        log_debug(ZONE, "retry limit reached for '%s' queue", rkey);
+                        log_debug(ZONE, "retry limit reached for '%.*s' queue", rkeylen, rkey);
                         q = NULL;
                     }
 
                     if (q != NULL && (npkt = jqueue_size(q)) > 0) {
                         conn_t retry;
 
-                        log_debug(ZONE, "retrying connection for '%s' queue", rkey);
-                        if (!out_route(out->s2s, rkey, &retry, 0)) {
+                        log_debug(ZONE, "retrying connection for '%.*s' queue", rkeylen, rkey);
+                        if (!out_route(out->s2s, rkey, rkeylen, &retry, 0)) {
                             log_debug(ZONE, "retry successful");
 
                             if (retry != NULL) {
                                 /* flush queue */
-                                out_flush_route_queue(out->s2s, rkey);
+                                out_flush_route_queue(out->s2s, rkey, rkeylen);
                             }
                         } else {
                             log_debug(ZONE, "retry failed");
 
                             /* bounce queue */
-                            out_bounce_route_queue(out->s2s, rkey, stanza_err_SERVICE_UNAVAILABLE);
+                            out_bounce_route_queue(out->s2s, rkey, rkeylen, stanza_err_SERVICE_UNAVAILABLE);
                         }
                     } else {
                         /* bounce queue */
-                        out_bounce_route_queue(out->s2s, rkey, stanza_err_SERVICE_UNAVAILABLE);
+                        out_bounce_route_queue(out->s2s, rkey, rkeylen, stanza_err_SERVICE_UNAVAILABLE);
                     }
                 } while(xhash_iter_next(out->routes));
             }
@@ -1243,6 +1355,7 @@ static int _out_mio_callback(mio_t m, mio_action_t a, mio_fd_t fd, void *data, v
 void send_dialbacks(conn_t out)
 {
   char *rkey;
+  int rkeylen;
 
   if (out->s2s->dns_bad_timeout > 0) {
       dnsres_t bad = xhash_get(out->s2s->dns_bad, out->key);
@@ -1258,8 +1371,8 @@ void send_dialbacks(conn_t out)
   if (xhash_iter_first(out->routes)) {
        log_debug(ZONE, "sending dialback packets for %s", out->key);
        do {
-            xhash_iter_get(out->routes, (const char **) &rkey, NULL);
-            _out_dialback(out, rkey);
+            xhash_iter_get(out->routes, (const char **) &rkey, &rkeylen, NULL);
+            _out_dialback(out, rkey, rkeylen);
           } while(xhash_iter_next(out->routes));
   }
 
@@ -1384,7 +1497,12 @@ static int _out_sx_callback(sx_t s, sx_event_t e, void *data, void *arg) {
                          strstr(sxe->specific, "not-authorized") ||   /* they do not want us there */
                          strstr(sxe->specific, "see-other-host") ||   /* we do not support redirections yet */
                          strstr(sxe->specific, "system-shutdown") ||  /* they are going down */
-                         strstr(sxe->specific, "unsupported-version") /* they do not support our stream version */
+                         strstr(sxe->specific, "policy-violation") || /* they do not want us there */
+                         strstr(sxe->specific, "remote-connection-failed") ||  /* the required remote entity is gone */
+                         strstr(sxe->specific, "unsupported-encoding") ||      /* they do not like our encoding */
+                         strstr(sxe->specific, "undefined-condition") ||       /* something bad happend */
+                         strstr(sxe->specific, "internal-server-error") ||     /* that server is broken */
+                         strstr(sxe->specific, "unsupported-version")          /* they do not support our stream version */
                         )))
             {
                 dnsres_t bad;
@@ -1516,6 +1634,7 @@ static void _out_result(conn_t out, nad_t nad) {
     int attr;
     jid_t from, to;
     char *rkey;
+    int rkeylen;
 
     attr = nad_find_attr(nad, 0, -1, "from", NULL);
     if(attr < 0 || (from = jid_new(NAD_AVAL(nad, attr), NAD_AVAL_L(nad, attr))) == NULL) {
@@ -1533,17 +1652,18 @@ static void _out_result(conn_t out, nad_t nad) {
     }
 
     rkey = s2s_route_key(NULL, to->domain, from->domain);
+    rkeylen = strlen(rkey);
 
     /* key is valid */
     if(nad_find_attr(nad, 0, -1, "type", "valid") >= 0) {
-        log_write(out->s2s->log, LOG_NOTICE, "[%d] [%s, port=%d] outgoing route '%s' is now valid%s", out->fd->fd, out->ip, out->port, rkey, (out->s->flags & SX_SSL_WRAPPER) ? ", TLS negotiated" : "");
+        log_write(out->s2s->log, LOG_NOTICE, "[%d] [%s, port=%d] outgoing route '%s' is now valid%s%s", out->fd->fd, out->ip, out->port, rkey, (out->s->flags & SX_SSL_WRAPPER) ? ", TLS negotiated" : "", out->s->compressed ? ", ZLIB compression enabled" : "");
 
         xhash_put(out->states, pstrdup(xhash_pool(out->states), rkey), (void *) conn_VALID);    /* !!! small leak here */
 
         log_debug(ZONE, "%s valid, flushing queue", rkey);
 
         /* flush the queue */
-        out_flush_route_queue(out->s2s, rkey);
+        out_flush_route_queue(out->s2s, rkey, rkeylen);
 
         free(rkey);
 
@@ -1568,7 +1688,7 @@ static void _out_result(conn_t out, nad_t nad) {
     sx_close(out->s);
 
     /* bounce queue */
-    out_bounce_route_queue(out->s2s, rkey, stanza_err_SERVICE_UNAVAILABLE);
+    out_bounce_route_queue(out->s2s, rkey, rkeylen, stanza_err_SERVICE_UNAVAILABLE);
 
     free(rkey);
 
@@ -1625,7 +1745,7 @@ static void _out_verify(conn_t out, nad_t nad) {
     attr = nad_find_attr(nad, 0, -1, "type", "valid");
     if(attr >= 0) {
         xhash_put(in->states, pstrdup(xhash_pool(in->states), rkey), (void *) conn_VALID);
-        log_write(in->s2s->log, LOG_NOTICE, "[%d] [%s, port=%d] incoming route '%s' is now valid%s", in->fd->fd, in->ip, in->port, rkey, (in->s->flags & SX_SSL_WRAPPER) ? ", TLS negotiated" : "");
+        log_write(in->s2s->log, LOG_NOTICE, "[%d] [%s, port=%d] incoming route '%s' is now valid%s%s", in->fd->fd, in->ip, in->port, rkey, (in->s->flags & SX_SSL_WRAPPER) ? ", TLS negotiated" : "", in->s->compressed ? ", ZLIB compression enabled" : "");
         valid = 1;
     } else {
         log_write(in->s2s->log, LOG_NOTICE, "[%d] [%s, port=%d] incoming route '%s' is now invalid", in->fd->fd, in->ip, in->port, rkey);
@@ -1668,13 +1788,14 @@ static void _out_verify(conn_t out, nad_t nad) {
 int out_bounce_domain_queues(s2s_t s2s, const char *domain, int err)
 {
   char *rkey;
+  int rkeylen;
   int pktcount = 0;
 
   if (xhash_iter_first(s2s->outq)) {
       do {
-          xhash_iter_get(s2s->outq, (const char **) &rkey, NULL);
-          if(s2s_route_key_match(NULL, domain, rkey))
-              pktcount += out_bounce_route_queue(s2s, rkey, err);
+          xhash_iter_get(s2s->outq, (const char **) &rkey, &rkeylen, NULL);
+          if(s2s_route_key_match(NULL, (char *) domain, rkey, rkeylen))
+              pktcount += out_bounce_route_queue(s2s, rkey, rkeylen, err);
       } while(xhash_iter_next(s2s->outq));
   }
 
@@ -1682,13 +1803,13 @@ int out_bounce_domain_queues(s2s_t s2s, const char *domain, int err)
 }
 
 /* bounce all packets in the queue for route */
-int out_bounce_route_queue(s2s_t s2s, char *rkey, int err)
+int out_bounce_route_queue(s2s_t s2s, char *rkey, int rkeylen, int err)
 {
   jqueue_t q;
   pkt_t pkt;
   int pktcount = 0;
 
-  q = xhash_get(s2s->outq, rkey);
+  q = xhash_getx(s2s->outq, rkey, rkeylen);
   if(q == NULL)
      return 0;
 
@@ -1707,7 +1828,7 @@ int out_bounce_route_queue(s2s_t s2s, char *rkey, int err)
   }
 
   /* delete queue and remove domain from queue hash */
-  log_debug(ZONE, "deleting out packet queue for %s", rkey);
+  log_debug(ZONE, "deleting out packet queue for %.*s", rkeylen, rkey);
   rkey = q->key;
   jqueue_free(q);
   xhash_zap(s2s->outq, rkey);
@@ -1719,13 +1840,14 @@ int out_bounce_route_queue(s2s_t s2s, char *rkey, int err)
 int out_bounce_conn_queues(conn_t out, int err)
 {
   char *rkey;
+  int rkeylen;
   int pktcount = 0;
 
   /* bounce queues for all domains handled by this connection - iterate through routes */
   if (xhash_iter_first(out->routes)) {
       do {
-          xhash_iter_get(out->routes, (const char **) &rkey, NULL);
-          pktcount += out_bounce_route_queue(out->s2s, rkey, err);
+          xhash_iter_get(out->routes, (const char **) &rkey, &rkeylen, NULL);
+          pktcount += out_bounce_route_queue(out->s2s, rkey, rkeylen, err);
       } while(xhash_iter_next(out->routes));
   }
 
@@ -1734,30 +1856,33 @@ int out_bounce_conn_queues(conn_t out, int err)
 
 void out_flush_domain_queues(s2s_t s2s, const char *domain) {
   char *rkey;
+  int rkeylen;
   char *c;
+  int c_len;
 
   if (xhash_iter_first(s2s->outq)) {
       do {
-          xhash_iter_get(s2s->outq, (const char **) &rkey, NULL);
-          c = strchr(rkey, '/');
+          xhash_iter_get(s2s->outq, (const char **) &rkey, &rkeylen, NULL);
+          c = memchr(rkey, '/', rkeylen);
           c++;
-          if (!strcmp(domain, c))
-              out_flush_route_queue(s2s, rkey);
+          c_len = rkeylen - (c - rkey);
+          if (strncmp(domain, c, c_len) == 0)
+              out_flush_route_queue(s2s, rkey, rkeylen);
       } while(xhash_iter_next(s2s->outq));
   }
 }
 
-void out_flush_route_queue(s2s_t s2s, char *rkey) {
+void out_flush_route_queue(s2s_t s2s, char *rkey, int rkeylen) {
     jqueue_t q;
     pkt_t pkt;
     int npkt, i, ret;
 
-    q = xhash_get(s2s->outq, rkey);
+    q = xhash_getx(s2s->outq, rkey, rkeylen);
     if(q == NULL)
         return;
 
     npkt = jqueue_size(q);
-    log_debug(ZONE, "flushing %d packets for '%s' to out_packet", npkt, rkey);
+    log_debug(ZONE, "flushing %d packets for '%.*s' to out_packet", npkt, rkeylen, rkey);
 
     for(i = 0; i < npkt; i++) {
         pkt = jqueue_pull(q);
@@ -1774,7 +1899,7 @@ void out_flush_route_queue(s2s_t s2s, char *rkey) {
 
     /* delete queue for route and remove route from queue hash */
     if (jqueue_size(q) == 0) {
-        log_debug(ZONE, "deleting out packet queue for '%s'", rkey);
+        log_debug(ZONE, "deleting out packet queue for '%.*s'", rkeylen, rkey);
         rkey = q->key;
         jqueue_free(q);
         xhash_zap(s2s->outq, rkey);
