@@ -29,9 +29,17 @@
  * !!! this blocks for every auth.
  */
 
+#define _XOPEN_SOURCE 500	// need this to get crypt()
 #include "c2s.h"
 
 #ifdef STORAGE_LDAP
+#ifdef HAVE_CRYPT
+#include <unistd.h>
+#endif
+
+#ifdef HAVE_SSL
+#include <openssl/rand.h>
+#endif
 
 #include <lber.h>
 #define LDAP_DEPRECATED 1
@@ -98,18 +106,27 @@ int _ldapfull_hash_init(); // call it before use of other stuff
 int _ldapfull_chk_hashed(moddata_t data, const char *scheme, int salted, const char *hash, const char *passwd);
 int _ldapfull_set_hashed(moddata_t data, const char *scheme, const char *prefix, int saltlen, const char *passwd, char *buf, int buflen);
 #endif
+#ifdef HAVE_CRYPT
+int _ldapfull_chk_crypt(moddata_t data, const char *scheme, int salted, const char *hash, const char *passwd);
+int _ldapfull_set_crypt(moddata_t data, const char *scheme, const char *prefix, int saltlen, const char *passwd, char *buf, int buflen);
+#endif
 int _ldapfull_chk_clear(moddata_t data, const char *scheme, int salted, const char *hash, const char *passwd);
 int _ldapfull_set_clear(moddata_t data, const char *scheme, const char *prefix, int saltlen, const char *passwd, char *buf, int buflen);
 
 int _ldapfull_check_passhash(moddata_t data, const char *hash, const char *passwd);
 int _ldapfull_set_passhash(moddata_t data, char *scheme_name, const char *passwd, char *buf, int buflen);
+static int _ldapfull_check_password_bind(authreg_t ar, char *username, char *realm, char password[LDAPFULL_PASSBUF_MAX]);
 
 ldapfull_pw_scheme _ldapfull_pw_schemas[] = {
 #ifdef HAVE_SSL
     { "sha", "sha1", "{SHA}", 0, _ldapfull_chk_hashed, _ldapfull_set_hashed },
     { "ssha", "sha1", "{SSHA}", 4, _ldapfull_chk_hashed, _ldapfull_set_hashed },
+#endif
+#ifdef HAVE_CRYPT
+    { "crypt", "crypt", "", 2, _ldapfull_chk_crypt, _ldapfull_set_crypt },
 #endif
     { "clear", "", "", 0, _ldapfull_chk_clear, _ldapfull_set_clear },
+    { "bind", "", "", 0, NULL, NULL },
     { NULL, NULL, NULL, 0, NULL, NULL }
 };
 
@@ -163,8 +180,6 @@ int _ldapfull_check_passhash(moddata_t data, const char *hash, const char *passw
 // must provide with buffer of sufficient length, or it will fail
 int _ldapfull_set_passhash(moddata_t data, char *scheme_name, const char *passwd, char *buf, int buflen) {
     int n;
-    int plen;
-    int nlen;
 
     if( ! passwd ) {
         log_write(data->ar->c2s->log,LOG_ERR,"_ldapfull_set_passhash: passwd is NULL");
@@ -350,9 +365,63 @@ int _ldapfull_set_hashed(moddata_t data, const char *scheme, const char *prefix,
 }
 #endif // HAVE_SSL
 
+#ifdef HAVE_CRYPT
+/** Check UNIX style crypt hashed password */
+int _ldapfull_chk_crypt(moddata_t data, const char *scheme, int salted, const char *hash, const char *passwd) {
+    const char *encrypted;
+    char salt[3];
+    if (strlen(hash) != 13) {
+        log_write(data->ar->c2s->log, LOG_ERR, "Invalid crypt hash length %d", strlen(hash));
+        return 0;
+    }
+    salt[0] = hash[0];
+    salt[1] = hash[1];
+    salt[2] = 0;
+    encrypted = crypt(passwd, salt);
+    return !strcmp(encrypted, hash);
+}
+
+int _ldapfull_set_crypt(moddata_t data, const char *scheme, const char *prefix, int saltlen, const char *passwd, char *buf, int buflen) {
+    const char *encrypted;
+    unsigned char salt[3];
+    static const char saltchars[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";
+    if ((saltlen != 2) || (buflen < 14)) {
+        log_write(data->ar->c2s->log, LOG_ERR, "Invalid crypt hash params");
+        return 0;
+    }
+#ifdef HAVE_SSL
+    if( !RAND_bytes(salt,saltlen) )
+        return 0;
+    salt[0] = saltchars[salt[0] % 64];
+    salt[1] = saltchars[salt[1] % 64];
+    salt[2] = 0;
+#else
+    /* Note: This is not a cryptographically secure random number generator */
+    salt[0] = saltchars[random() % 64];
+    salt[1] = saltchars[random() % 64];
+    salt[2] = 0;
+#endif
+    encrypted = crypt(passwd, salt);
+    strncpy(buf, encrypted, buflen);
+    buf[buflen-1] = 0;
+    return 1;
+}
+#endif // HAVE_CRYPT
+
+
+/** Makes a copy of enough LDAP data in order to establish a second connection */
+static void copy_ldap_config(moddata_t from, moddata_t to)
+{
+    memset(to, 0, sizeof(struct moddata_st));
+    to->uri = from->uri;
+    to->ar = from->ar;
+}
+
 int _ldapfull_hash_init() {
 #ifdef HAVE_SSL
     OpenSSL_add_all_digests();
+#else
+    srandom(time(NULL) ^ getpid());
 #endif
     return 1;
 }
@@ -515,22 +584,53 @@ retry:
 }
 
 /** do we have this user? */
-static int _ldapfull_user_exists(authreg_t ar, char *username, char *realm)
+static int _ldapfull_find_user_dn(moddata_t data, char *username, char *realm, char **dn)
 {
-    moddata_t data = (moddata_t) ar->private;
-    char *dn;
-
+    *dn = NULL;
     if(_ldapfull_connect_bind(data))
-        return 0;
+        return 0; // error
 
     log_debug(ZONE, "checking existance of %s", username);
     
-    if( dn = _ldapfull_search(data, realm, username) ) {
-        ldap_memfree(dn);
-        return 1;
-    } else {
-        return 0;
+    *dn = _ldapfull_search(data, realm, username);
+    return *dn != NULL;
+}
+
+/** do we have this user? */
+static int _ldapfull_user_exists(authreg_t ar, char *username, char *realm)
+{
+    char *dn;
+    if (_ldapfull_find_user_dn((moddata_t) ar->private, username, realm, &dn)) {
+	ldap_memfree(dn);
+	return 1;
     }
+    return 0;
+}
+
+/** This method determines the DN of the user and does a new simple bind of the LDAP
+server. If the server allows it, the user has been authenticated.
+*/
+static int _ldapfull_check_password_bind(authreg_t ar, char *username, char *realm, char password[LDAPFULL_PASSBUF_MAX])
+{
+    moddata_t data = (moddata_t) ar->private;
+    struct moddata_st bind_data;
+    int invalid;
+    char *dn;
+
+    if (!_ldapfull_find_user_dn(data, username, realm, &dn)) {
+        log_debug(ZONE, "User %s not found", username);
+	return 0;
+    }
+
+    /* Try logging in to the LDAP server as this user's DN */
+    copy_ldap_config(data, &bind_data);
+    bind_data.binddn = dn;
+    bind_data.bindpw = password;
+    invalid = _ldapfull_connect_bind(&bind_data);
+    if (!invalid)
+        _ldapfull_unbind(&bind_data);
+    ldap_memfree(dn);
+    return invalid;
 }
 
 // get password from jabberPassword attribute
@@ -654,6 +754,11 @@ static int _ldapfull_check_password(authreg_t ar, char *username, char *realm, c
     if(password[0] == '\0')
         return 1;
 
+    /* The bind scheme doesn't need the password read first, so short circuit
+       the whole passhash scheme */
+    if (!strcmp(data->pwscheme, "bind"))
+        return _ldapfull_check_password_bind(ar, username, realm, password);
+
     if( _ldapfull_get_password(ar,username,realm,buf) != 0  ) {
         return 1;
     }
@@ -704,7 +809,7 @@ DLLEXPORT int ar_init(authreg_t ar)
     basedn = config_get(ar->c2s->config, "authreg.ldapfull.basedn");
     if(basedn == NULL)
     {
-        log_write(ar->c2s->log, LOG_ERR, "ldap: no basedns specified in config file");
+        log_write(ar->c2s->log, LOG_ERR, "ldap: no basedn specified in config file");
         return 1;
     }
 
@@ -799,4 +904,4 @@ DLLEXPORT int ar_init(authreg_t ar)
     return 0;
 }
 
-#endif
+#endif // STORAGE_LDAP
